class DataPlotter(object):
    def plot_batch(self, train_set_inputs, train_set_output, synthetic_output, epoch, output_dir):
        raise NotImplementedError("plot_batch method must be implemented")